from typing import AsyncIterator, Iterator

from example.services import (
    AsyncDatabase,
    AsyncDatabaseEngine,
    Database,
    DatabaseEngine,
    Settings,
    UserManager,
)
from pure_di import AsyncProvider, Provider

provider = Provider()


@provider.include_factory
def provide_settings() -> Settings:
    return Settings("db_uri")


@provider.include_factory(service_type=DatabaseEngine)
def provide_engine(settings: Settings) -> Iterator[DatabaseEngine]:
    with DatabaseEngine.open(settings.db_uri) as engine:
        yield engine


provider.include_factory(Database, service_type=Database)

async_provider = AsyncProvider()


@async_provider.include_factory(service_type=AsyncDatabaseEngine)
async def provide_async_engine(
    settings: Settings,
) -> AsyncIterator[AsyncDatabaseEngine]:
    async with AsyncDatabaseEngine.open(settings.db_uri) as engine:
        yield engine


async_provider.include_factory(AsyncDatabase, service_type=AsyncDatabase)


@async_provider.include_factory
async def provide_user_manager(
    sync_db: Database, async_db: AsyncDatabase
) -> UserManager:
    await async_db.find()
    return UserManager(sync_db, async_db)



import functools
import inspect
from typing import Callable

from pure_di import AsyncProvider, Provider
from pure_di._provider import _BaseProvider


def pytest_provide(provider: Provider):
    def decorate(initial_test: Callable):
        @functools.wraps(initial_test)
        def new_test_func(*args, **kwargs):
            with provider.build() as container:
                return container.partial_solve(initial_test)(*args, **kwargs)

        new_test_func.__signature__ = _fix_signature(  # type: ignore[attr-defined]
            provider, initial_test
        )
        _inherit_pytest_mark(new_test_func, initial_test)
        return new_test_func

    return decorate


def pytest_provide_async(provider: AsyncProvider):
    def decorate(initial_test: Callable):
        @functools.wraps(initial_test)
        async def new_test_func(*args, **kwargs):
            async with provider.build_async() as container:
                return await container.partial_solve(initial_test)(*args, **kwargs)

        new_test_func.__signature__ = _fix_signature(  # type: ignore[attr-defined]
            provider, initial_test
        )
        _inherit_pytest_mark(new_test_func, initial_test)
        return new_test_func

    return decorate


def _fix_signature(provider: _BaseProvider, initial_test: Callable):
    initial_sig = inspect.signature(initial_test)
    parameters = list(initial_sig.parameters.values())
    new_parameters = [param for param in parameters if param.annotation not in provider]
    return initial_sig.replace(parameters=new_parameters)


def _inherit_pytest_mark(new_test_func: Callable, initial_test_func):
    pytestmark = getattr(initial_test_func, "pytestmark", None)
    if pytestmark is not None:
        new_test_func.pytestmark = pytestmark  # type: ignore[attr-defined]



import pytest

SYNC_FIXTURE_VALUE = object()
ASYNC_FIXTURE_VALUE = object()


@pytest.fixture()
def sync_fixture():
    return SYNC_FIXTURE_VALUE


@pytest.fixture()
async def async_fixture():
    return ASYNC_FIXTURE_VALUE



from __future__ import annotations

import functools
import inspect
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import (
    AsyncContextManager,
    AsyncIterator,
    Callable,
    ContextManager,
    Generic,
    Type,
)

from pure_di._types import Service, TService


@dataclass
class _Inspector(Generic[TService]):
    initial_callable: Callable
    signature: inspect.Signature = field(init=False)

    def __post_init__(self):
        self.signature = inspect.signature(self.initial_callable)

    def get_argument_annotations(self) -> list[Type[Service]]:
        for parameter in self.signature.parameters.values():
            if parameter.annotation is inspect.Signature.empty:
                raise TypeError(
                    f"{self.initial_callable.__name__} parameter {parameter}"
                    f" does not have annotation"
                )
        return [
            parameter.annotation for parameter in self.signature.parameters.values()
        ]

    def get_return_annotation(self) -> Type[TService]:
        # ToDo handle types like ContextManager[T]
        # ToDo: try handle ForwardRef
        if self.signature.return_annotation is inspect.Signature.empty:
            raise TypeError(
                f"{self.initial_callable.__name__} does not have return annotation"
            )
        return self.signature.return_annotation

    def wrap_to_async_context_manager(
        self,
    ) -> Callable[..., AsyncContextManager[TService]]:
        if inspect.iscoroutinefunction(self.initial_callable):

            @asynccontextmanager
            @functools.wraps(self.initial_callable)
            async def _build(*args: Service) -> AsyncIterator[TService]:
                yield await self.initial_callable(*args)

        elif inspect.isasyncgenfunction(self.initial_callable):
            _build = asynccontextmanager(self.initial_callable)

        elif inspect.isgeneratorfunction(self.initial_callable):
            _sync_context_build = contextmanager(self.initial_callable)

            @asynccontextmanager
            @functools.wraps(self.initial_callable)
            async def _build(*args: Service) -> AsyncIterator[TService]:
                with _sync_context_build(*args) as result:
                    yield result

        else:

            @asynccontextmanager
            @functools.wraps(self.initial_callable)
            async def _build(*args: Service) -> AsyncIterator[TService]:
                yield self.initial_callable(*args)

        return _build

    def wrap_to_sync_context_manager(self) -> Callable[..., ContextManager[TService]]:
        if inspect.isgeneratorfunction(self.initial_callable):
            _build = contextmanager(self.initial_callable)
        else:

            @contextmanager
            @functools.wraps(self.initial_callable)
            def _build(*args: Service):
                yield self.initial_callable(*args)

        return _build



"""Simple yet powerful dependency injection framework!"""

from pure_di._container import Container
from pure_di._provider import AsyncProvider, Provider

__version__ = "0.0.0"
__all__ = ["Provider", "AsyncProvider", "Container"]



from enum import Enum, auto
from typing import Any, Dict, Type, TypeVar

from typing_extensions import TypeAlias

Service: TypeAlias = Any
ContainerDict = Dict[Type[Service], Service]

TService = TypeVar("TService")


class _Empty(Enum):
    empty = auto()



from __future__ import annotations

import functools
import inspect
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from typing import AsyncIterator, Callable, Iterator, Optional, Type

from pure_di._factory import _AsyncFactory, _BaseFactory, _SyncFactory
from pure_di._service_dict import ServiceDict
from pure_di._types import Service, TService


class _PrivateContainer(ServiceDict[Service]):
    @asynccontextmanager
    async def solve_async_factory(
        self, factory: _AsyncFactory[TService]
    ) -> AsyncIterator[TService]:
        deps = [self.data[dep] for dep in factory.dependencies]
        async with factory.async_build(*deps) as service:
            yield service

    @contextmanager
    def solve_sync_factory(self, factory: _SyncFactory[TService]) -> Iterator[TService]:
        deps = [self.data[dep] for dep in factory.dependencies]
        with factory.sync_build(*deps) as service:
            yield service

    def check_factory_solvable(self, factory: _BaseFactory[Service]) -> bool:
        return all([dep in self.data for dep in factory.dependencies])

    def clone_and_include(
        self, service_type: Type[Service], service: Service
    ) -> "_PrivateContainer":
        return self.merge(ServiceDict({service_type: service}))


@dataclass
class Container:
    _private: _PrivateContainer

    def get(self, service_type: Type[TService]) -> Optional[TService]:
        return self._private.get(service_type)

    def keys(self) -> set[Type[Service]]:
        return set(self._private.keys())

    def __getitem__(self, service_type: Type[TService]) -> TService:
        service = self.get(service_type)
        if service is None:
            raise KeyError()  # ToDo: better errors
        return service

    def __repr__(self):
        return f"{type(self).__name__}({repr(self._private)})"

    def partial_solve(self, callable_: Callable) -> Callable:
        sig = inspect.signature(callable_)
        kwargs = {
            p.name: self._private[p.annotation]
            for p in sig.parameters.values()
            if p.annotation in self._private
        }
        return functools.partial(callable_, **kwargs)



import pytest

from example.factories import async_provider, provider
from example.services import (
    AsyncDatabase,
    AsyncDatabaseEngine,
    Database,
    DatabaseEngine,
    Settings,
    UserManager,
)
from pure_di import AsyncProvider, Container, Provider

pytestmark = pytest.mark.anyio


def test_merge_creates_new_provider():
    assert all(
        [
            provider is not provider | async_provider,
            async_provider is not provider | async_provider,
            provider | async_provider is not async_provider | provider,
            provider | async_provider is not provider | async_provider,
        ]
    )


def test_provider_types():
    assert type(provider) is Provider
    assert type(async_provider) is AsyncProvider
    assert type(provider | async_provider) is AsyncProvider

    assert type(Provider.from_dict({})) is Provider
    assert type(provider | Provider.from_dict({})) is Provider


def test_container_repr():
    with Provider().build() as container:
        assert str(container) == "Container({})"


def test_provider_copy():
    copy_provider = provider.copy()
    assert type(copy_provider) is Provider

    @copy_provider.include_factory
    def copy_settings() -> Settings:
        return Settings("copy")

    with copy_provider.build() as container:
        assert container[Database].find() == "copy"

    with provider.build() as container:
        assert container[Database].find() == "db_uri"


def test_example_sync_provider():
    with provider.build() as container:
        assert type(container) is Container
        with pytest.raises(KeyError):
            _ = container[AsyncDatabase]
        assert container.keys() == {
            Settings,
            DatabaseEngine,
            Database,
        }
        assert container[Database].find() == "db_uri"

        def to_be_solved(settings: Settings):
            return settings.db_uri

        assert container.partial_solve(to_be_solved)() == "db_uri"


async def test_example_async_provider():
    async with (provider | async_provider).build_async() as container:
        assert type(container) is Container
        assert container.keys() == {
            Settings,
            DatabaseEngine,
            AsyncDatabaseEngine,
            Database,
            AsyncDatabase,
            UserManager,
        }

        db = container[Database]
        async_db = container[AsyncDatabase]
        assert db.find() == "db_uri"
        assert await async_db.find() == "db_uri"
        assert await container[UserManager].create_user() == "user"

        async def to_be_solved(user_manager: UserManager):
            return await user_manager.create_user()

        assert await container.partial_solve(to_be_solved)() == "user"

    with pytest.raises(RuntimeError):
        db.find()

    with pytest.raises(RuntimeError):
        await async_db.find()


async def test_override_settings():
    with (
        provider | Provider.from_dict({Settings: Settings("other_db_uri")})
    ).build() as container:
        assert container[Database].find() == "other_db_uri"

    async with (
        provider
        | async_provider
        | Provider.from_dict({Settings: Settings("other_db_uri")})
    ).build_async() as container:
        assert await container[AsyncDatabase].find() == "other_db_uri"


def test_provider_from_factory():
    @Provider.from_factory
    def provide_int() -> int:
        return 1

    @Provider.from_factory(service_type=float)
    def provide_float() -> int:
        return 2

    @Provider.from_factory()
    def provide_str(a: int, b: float) -> str:
        return str(a + b)

    provider_ = provide_int | provide_float | provide_str
    assert type(provider_) is Provider
    with (provide_int | provide_float | provide_str).build() as container:
        assert container[str] == "3"


async def test_unsolvable():
    with pytest.raises(RuntimeError):
        async with async_provider.build_async() as _:
            ...



import pytest

from pure_di import AsyncProvider, Provider


def test_no_return_annotation():
    with pytest.raises(TypeError):

        @Provider.from_factory
        def provide_something():
            return "something"

    with pytest.raises(TypeError):

        @Provider.from_factory
        def provide_something_x(x) -> str:
            return "something" + x


def test_cant_convert_to_factory():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        Provider.from_factory(3)

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        AsyncProvider.from_factory(3)



from __future__ import annotations

from collections import UserDict
from typing import TYPE_CHECKING, Generic, Type, TypeVar

from typing_extensions import Self

from pure_di._types import Service

ServiceDictValue = TypeVar("ServiceDictValue")

if TYPE_CHECKING:
    Base = UserDict[Type[Service], ServiceDictValue]  # pragma: no cover
else:
    Base = UserDict


class ServiceDict(Generic[ServiceDictValue], Base):
    def is_empty(self) -> bool:
        return not self.data

    def merge(
        self, other: ServiceDict[ServiceDictValue]
    ) -> Self:  # type: ignore[valid-type]
        new_data = {}
        new_data.update(self.data)
        new_data.update(other.data)
        # noinspection PyArgumentList
        return type(self)(new_data)

    def remove_keys_from(self, other: ServiceDict) -> Self:  # type: ignore[valid-type]
        for key in other.keys():
            self.pop(key, None)  # type: ignore[arg-type]
        return self

    def copy_data_with_removed(
        self, remove_service_type: Type[Service]
    ) -> dict[Type[Service], ServiceDictValue]:
        new_data = self.data.copy()
        del new_data[remove_service_type]
        return new_data

    def copy(self) -> Self:  # type: ignore[valid-type]
        # noinspection PyArgumentList
        return type(self)(self.data.copy())



from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import (
    AsyncIterator,
    Callable,
    Generic,
    Iterator,
    Type,
    TypeVar,
    cast,
    overload,
)

from pure_di._container import Container, _PrivateContainer
from pure_di._factory import _AsyncFactory, _SyncFactory
from pure_di._factory_group import (
    _AsyncFactoryGroup,
    _BaseFactoryGroup,
    _FactoryType,
    _SyncFactoryGroup,
)
from pure_di._types import Service, _Empty
from pure_di.utils import get_generic_first_type

_FactoryGroupType = TypeVar("_FactoryGroupType", bound=_BaseFactoryGroup)


@dataclass
class _BaseProvider(Generic[_FactoryGroupType, _FactoryType]):
    _factories: _FactoryGroupType
    _container: _PrivateContainer = field(default_factory=_PrivateContainer)

    @classmethod
    def from_container(cls, container: _PrivateContainer):
        factory_group_type: Type[_FactoryGroupType] = get_generic_first_type(cls)
        return cls(factory_group_type({}), container)

    @classmethod
    def from_dict(
        cls, service_dict: dict[Type[Service], Service]
    ) -> _BaseProvider:  # Self
        return cls.from_container(_PrivateContainer(service_dict))

    @classmethod
    @overload
    def from_factory(
        cls,
        *,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ) -> Callable[..., _BaseProvider]:  # Self
        ...

    @classmethod
    @overload
    def from_factory(
        cls,
        initial_callable: Callable,
        *,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ) -> _BaseProvider:
        ...

    @classmethod
    def from_factory(
        cls,
        initial_callable: Callable | _Empty = _Empty.empty,
        *,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ):
        def decorator(
            initial_callable_: Callable,
        ) -> _BaseProvider:
            return cls.from_dict({}).include_factory(
                initial_callable_, service_type=service_type
            )

        if initial_callable is _Empty.empty:
            return decorator
        return decorator(initial_callable)

    def _find_factory_to_solve(self) -> tuple[Type[Service], _FactoryType]:
        for service_type, factory in self._factories.items():
            if self._container.check_factory_solvable(factory):
                return service_type, factory
        else:
            raise RuntimeError(
                f"Can not solve any factory {self}"
            )  # ToDo: better errors

    def _clone_with_new_service(
        self, service_type: Type[Service], service: Service
    ) -> _BaseProvider:
        next_factories = self._factories.clone_and_remove(service_type)
        next_container = self._container.clone_and_include(service_type, service)
        # noinspection PyArgumentList
        return type(self)(next_factories, next_container)

    def _merge_self(self, other: _BaseProvider) -> _BaseProvider:
        # noinspection PyArgumentList
        return type(self)(
            _factories=cast(
                _FactoryGroupType, self._factories.merge(other._factories)
            ).remove_keys_from(other._container),
            _container=cast(
                _PrivateContainer, self._container.merge(other._container)
            ).remove_keys_from(other._factories),
        )

    @overload
    def include_factory(
        self, *, service_type: Type[Service] | _Empty = _Empty.empty
    ) -> Callable[[Callable], _BaseProvider]:  # Self
        ...

    @overload
    def include_factory(
        self,
        initial_callable: Callable,
        *,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ) -> _BaseProvider:  # Self
        ...

    def include_factory(
        self,
        initial_callable: Callable | _Empty = _Empty.empty,
        *,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ):
        def decorator(initial_callable_: Callable) -> _BaseProvider:  # Self
            self._factories.include_factory(initial_callable_, service_type)
            self._container.pop(service_type, None)  # type: ignore[arg-type]
            return self

        if initial_callable is _Empty.empty:
            return decorator
        return decorator(initial_callable)

    def _to_async_provider(self) -> AsyncProvider:
        return AsyncProvider(
            _factories=self._factories.to_async_factory_group(),
            _container=self._container,
        )

    def __contains__(self, service_type: Type[Service]) -> bool:
        return service_type in self._factories or service_type in self._container

    def copy(self) -> _BaseProvider:  # Self
        # noinspection PyArgumentList
        return type(self)(self._factories.copy(), self._container.copy())


@dataclass
class AsyncProvider(_BaseProvider[_AsyncFactoryGroup, _AsyncFactory]):
    _factories: _AsyncFactoryGroup = field(default_factory=_AsyncFactoryGroup)

    @asynccontextmanager
    async def build_async(self) -> AsyncIterator[Container]:
        if self._factories.is_empty():
            yield Container(self._container)
            return

        async with self._async_solve_factory() as next_provider:
            async with next_provider.build_async() as result_container:
                yield result_container

    @asynccontextmanager
    async def _async_solve_factory(
        self,
    ) -> AsyncIterator["AsyncProvider"]:
        service_type, factory = self._find_factory_to_solve()
        async with self._container.solve_async_factory(factory) as new_service:
            yield cast(
                AsyncProvider, self._clone_with_new_service(service_type, new_service)
            )

    def _to_async_provider(self) -> AsyncProvider:
        return self

    def __or__(self, other: Provider | AsyncProvider) -> AsyncProvider:
        return cast(AsyncProvider, self._merge_self(other._to_async_provider()))


@dataclass
class Provider(_BaseProvider[_SyncFactoryGroup, _SyncFactory]):
    _factories: _SyncFactoryGroup = field(default_factory=_SyncFactoryGroup)

    @contextmanager
    def build(self) -> Iterator[Container]:
        if self._factories.is_empty():
            yield Container(self._container)
            return

        with self._sync_solve_factory() as next_provider:
            with next_provider.build() as result_container:
                yield result_container

    @contextmanager
    def _sync_solve_factory(self) -> Iterator[Provider]:
        service_type, factory = self._find_factory_to_solve()
        with self._container.solve_sync_factory(factory) as new_service:
            yield cast(
                Provider, self._clone_with_new_service(service_type, new_service)
            )

    @overload
    def __or__(self, other: Provider) -> Provider:
        ...

    @overload
    def __or__(self, other: AsyncProvider) -> AsyncProvider:
        ...

    def __or__(self, other: Provider | AsyncProvider):
        if isinstance(other, Provider):
            return self._merge_self(other)
        return self._to_async_provider()._merge_self(other._to_async_provider())



from __future__ import annotations

from typing import Callable, Type, TypeVar

from typing_extensions import Self

from pure_di._factory import _AsyncFactory, _BaseFactory, _SyncFactory
from pure_di._service_dict import ServiceDict
from pure_di._types import Service, _Empty
from pure_di.utils import get_generic_first_type

_FactoryType = TypeVar("_FactoryType", bound=_BaseFactory)


class _BaseFactoryGroup(ServiceDict[_FactoryType]):
    def clone_and_remove(
        self, service_type: Type[Service]
    ) -> Self:  # type: ignore[valid-type]
        return type(self)(self.copy_data_with_removed(service_type))

    def include_factory(
        self,
        initial_callable: Callable,
        service_type: Type[Service] | _Empty = _Empty.empty,
    ):
        factory_type: Type[_BaseFactory] = get_generic_first_type(self)
        factory = factory_type.from_callable(initial_callable, service_type)
        self.data[factory.service_type] = factory

    def to_async_factory_group(self) -> "_AsyncFactoryGroup":
        return _AsyncFactoryGroup(
            {
                service_type: factory.to_async_factory()
                for service_type, factory in self.items()
            }
        )


class _AsyncFactoryGroup(_BaseFactoryGroup[_AsyncFactory]):
    ...


class _SyncFactoryGroup(_BaseFactoryGroup[_SyncFactory]):
    ...



from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from typing import AsyncIterator, Iterator


@dataclass
class Settings:
    db_uri: str


@dataclass
class DatabaseEngine:
    db_uri: str
    is_opened: bool = False

    @classmethod
    @contextmanager
    def open(cls, db_uri: str) -> Iterator["DatabaseEngine"]:
        engine = cls(db_uri)
        engine.is_opened = True
        yield engine
        engine.is_opened = False

    def find(self):
        if self.is_opened:
            return self.db_uri
        raise RuntimeError("closed")


@dataclass
class AsyncDatabaseEngine:
    db_uri: str
    is_opened: bool = False

    @classmethod
    @asynccontextmanager
    async def open(cls, db_uri: str) -> AsyncIterator["AsyncDatabaseEngine"]:
        engine = cls(db_uri)
        engine.is_opened = True
        yield engine
        engine.is_opened = False

    async def find(self):
        if self.is_opened:
            return self.db_uri
        raise RuntimeError("closed")


@dataclass
class Database:
    engine: DatabaseEngine

    def find(self):
        return self.engine.find()


@dataclass
class AsyncDatabase:
    engine: AsyncDatabaseEngine

    async def find(self):
        return await self.engine.find()


@dataclass
class UserManager:
    sync_db: Database
    async_db: AsyncDatabase

    async def create_user(self):
        assert type(self.sync_db) is Database
        assert type(self.async_db) is AsyncDatabase
        return "user"



import nox


@nox.session(python=["3.8", "3.9", "3.10", "3.11"])
def tests(session: nox.Session):
    session.install(".")
    session.install("-r", "requirements/test.txt")
    session.run("pytest")



import pytest

from example.factories import async_provider, provider
from example.services import Settings, UserManager
from pure_di.pytest import pytest_provide, pytest_provide_async
from tests.conftest import ASYNC_FIXTURE_VALUE, SYNC_FIXTURE_VALUE


@pytest_provide(provider)
def test_provide_by_type(settings: Settings, sync_fixture):
    assert settings.db_uri == "db_uri"
    assert sync_fixture == SYNC_FIXTURE_VALUE


@pytest_provide_async(provider | async_provider)
@pytest.mark.anyio
async def test_provide_by_type_async(user_manager: UserManager, async_fixture):
    assert await user_manager.create_user() == "user"
    assert async_fixture == ASYNC_FIXTURE_VALUE


@pytest.mark.anyio
async def test_mark():
    assert test_provide_by_type_async.pytestmark == test_mark.pytestmark



from __future__ import annotations

from dataclasses import dataclass, field
from typing import AsyncContextManager, Callable, ContextManager, Generic, Type

from pure_di._inspector import _Inspector
from pure_di._types import Service, TService, _Empty


@dataclass
class _BaseFactory(Generic[TService]):
    initial_callable: Callable = field(repr=False)
    service_type: Type[TService]
    dependencies: list[Type[Service]]
    _inspector: _Inspector = field(repr=False)

    @classmethod
    def from_callable(
        cls,
        initial_callable: Callable,
        service_type: Type[TService] | _Empty = _Empty.empty,
    ):
        inspector: _Inspector = _Inspector(initial_callable)
        guess_service_type: Type[TService] = (
            service_type
            if service_type is not _Empty.empty
            else inspector.get_return_annotation()
        )
        return cls(
            initial_callable=initial_callable,
            dependencies=inspector.get_argument_annotations(),
            service_type=guess_service_type,
            _inspector=inspector,
        )

    def to_async_factory(self) -> "_AsyncFactory":
        return _AsyncFactory(
            initial_callable=self.initial_callable,
            dependencies=self.dependencies,
            service_type=self.service_type,
            _inspector=self._inspector,
        )


@dataclass
class _AsyncFactory(_BaseFactory[TService]):
    async_build: Callable[..., AsyncContextManager[TService]] = field(init=False)

    def __post_init__(self):
        self.async_build = self._inspector.wrap_to_async_context_manager()


@dataclass
class _SyncFactory(_BaseFactory[TService]):
    sync_build: Callable[..., ContextManager[TService]] = field(init=False)

    def __post_init__(self):
        self.sync_build = self._inspector.wrap_to_sync_context_manager()
